/*******************************************\

  OpenGL code for Stroke Based Rendering
  Bump mapped oil paint strokes
  
  (c) John Collomosse 2005

  Created: JPC 22/6/05

  Extracted from "Empathic Painting" NPAR 2006 for 
  EG2011 NPR tutorial (Collomosse and Kyprianidis)

  Released under GPLv3
  http://www.gnu.org/licenses/gpl.html  

\*******************************************/

#define WIN32_LEAN_AND_MEAN (1)

// These OpenCV dependencies are used for loading image files
// preferred over win32 calls for portability to Linux etc.
// but easily substituted.
#include "cxcore.h"
#include "highgui.h"

#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#include "windows.h" // comment out for Linux compilation
#include "GL/gl.h"
#include "GL/glu.h"
#include "GL/glut.h"



#include "sysparams.h"
#include "painting.h"
#include "Extensions.h"
#include "extraGLstuff.h"

typedef float VECTOR3[3];
STROKE* painting=NULL;
int uniqueID=0;

/* Min/max functions - comment these definitions if already available */
int min(int a, int b) { return a<b?a:b; }
float min(float a, float b) { return a<b?a:b; }
double min(double a, double b) { return a<b?a:b; }
int max(int a, int b) { return a>b?a:b; }
float max(float a, float b) { return a>b?a:b; }
double max(double a, double b) { return a>b?a:b; }



/******* PROTOTYPES ******/
void	GLdisplay	(void);
int		main		(int, char**);
void	GLreshape	(int, int);
void	GLkeyEventSpecial (int, int, int);
void	GLkeyEventStandard (unsigned char, int, int);
void	GLidleDispatch (void);
void	uploadTexture(IplImage*, IplImage*, int);
void	GLinit (int,char**);
void	plotSplines(void);
void	freePainting (void);
void mouseCallback(int event, int x, int y, int flags);
float getSplineLength4pt(CvPoint2D32f*);
void trackbarRedisplay(int dummy);
int parseComms(int*);
int searchFor (unsigned char*, int, unsigned char*, int);
void convRGB2HSV (float*, float*, float*);
void convHSV2RGB (float*, float*, float*);

/******* GLOBAL VARIABLES ******/
int CANVAS_HEIGHT, CANVAS_WIDTH;
int textureNames[NUM_TEX];
int isFullScreen=0;
IplImage* sor;
VECTOR3 light_position;
int IMAGE_HEIGHT;
int IMAGE_WIDTH;

double* costheta;
double* sintheta;

double* triangletheta;
double* derivtriangletheta;



RENDERPARAMS renderparams;



/** Function called once when system is calibrated (but before mainIdleLoop calls start) **/
int main(int argc, char** argv) {

	int i;
	double nearpeak;


	renderparams.undulation_amp=renderparams.undulation_freq=0.5;
	renderparams.jaggedness=0;
	renderparams.dampen_bgstrokes=1;
	renderparams.bgcoverage=0;
	renderparams.colpleasant=0;
	renderparams.colarousal=0;
	renderparams.maxstrokeang=0.56f;
	renderparams.maxstrokelen=1.0f;



	rng_state=cvRNG((unsigned)time(NULL));


	IMAGE_WIDTH=800;
	IMAGE_HEIGHT=600;


	GLinit(argc,argv);

	/* build lookup table to accelerate SIN and COS */
	printf("Building trig. lookup tables...");
	sintheta=(double*)calloc(360*TRIG_LOOKUP_RESOLUTION,sizeof(double));
	costheta=(double*)calloc(360*TRIG_LOOKUP_RESOLUTION,sizeof(double));
	triangletheta=(double*)calloc(360*TRIG_LOOKUP_RESOLUTION,sizeof(double));
	derivtriangletheta=(double*)calloc(360*TRIG_LOOKUP_RESOLUTION,sizeof(double));
	for (i=0; i<360*TRIG_LOOKUP_RESOLUTION; i++) {
		sintheta[i]=sin(DEG_TO_RAD*((double)i/TRIG_LOOKUP_RESOLUTION));
		costheta[i]=cos(DEG_TO_RAD*((double)i/TRIG_LOOKUP_RESOLUTION));
		if ((double)i/TRIG_LOOKUP_RESOLUTION <=90) {
			triangletheta[i]=(((double)i/TRIG_LOOKUP_RESOLUTION)/90.0);
			derivtriangletheta[i]=1.0/sqrt(2.0);
		}
		else if ((double)i/TRIG_LOOKUP_RESOLUTION <=270) {
			triangletheta[i]=(1.0-((((double)i/TRIG_LOOKUP_RESOLUTION)-90.0)/180.0)*2.0);
			derivtriangletheta[i]=-1.0/sqrt(2.0);
		}
		else {
			triangletheta[i]=((((double)i/TRIG_LOOKUP_RESOLUTION)-270.0)/90.0)-1.0;
			derivtriangletheta[i]=1.0/sqrt(2.0);
		}
		nearpeak=fabs(((double)i/TRIG_LOOKUP_RESOLUTION)-90.0);
		nearpeak=min(nearpeak,fabs(((double)i/TRIG_LOOKUP_RESOLUTION)-180.0));
		nearpeak=min(nearpeak,fabs(((double)i/TRIG_LOOKUP_RESOLUTION)-270.0));
		nearpeak=min(nearpeak,fabs(((double)i/TRIG_LOOKUP_RESOLUTION)-360.0));
		if (nearpeak<5.0) { // hack to fix discontinuity issues with the triangular function - smooth caps frmo cosine
			derivtriangletheta[i]=(1.0-(nearpeak/5.0))*derivtriangletheta[i] + (nearpeak/5.0)*costheta[i];
		}
	}
	printf("OK\n");


	buildTestPainting();


	// pass control to openGL render loop	
	printf("Starting OpenGL render loop...\n");
	glutMainLoop();

	return 0; // warning eater	
	
}



void GLinit(int argc, char** argv) {

	IplImage* scratch;
	IplImage* brushTexOpacity;
	IplImage* brushTexMask;
	IplImage* brushTexBump;
	IplImage* greytmp;
	int miplevel;
	
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
	glutInitWindowSize  ( 400, 400 ); // If glutFullScreen wasn't called this is the window size
	CANVAS_HEIGHT=400; CANVAS_WIDTH=400;


	/** GL VARIABLES **/
	glShadeModel(GL_SMOOTH);                        // Enable flat shading

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);


	
	glutCreateWindow    ( "Painterly Output" ); // Window Title (argv[0] for current directory as title)

	glutDisplayFunc     (GLdisplay);  // Matching Earlier Functions To Their Counterparts
	glutReshapeFunc     (GLreshape);
	glutKeyboardFunc    (GLkeyEventStandard);
	glutSpecialFunc     (GLkeyEventSpecial);
	glutIdleFunc     (GLidleDispatch);
    //glutFullScreen ( ); // Go Into Full Screen Mode
//	glutSetCursor(GLUT_CURSOR_NONE); // no mouse cursor

	glGenTextures(NUM_TEX,(unsigned int*)textureNames);

	brushTexOpacity=cvCreateImage(cvSize(TEXTURE_RESOLUTION,TEXTURE_RESOLUTION),8,3);
	brushTexMask=cvCreateImage(cvSize(TEXTURE_RESOLUTION,TEXTURE_RESOLUTION),8,1);
	brushTexBump=cvCreateImage(cvSize(TEXTURE_RESOLUTION,TEXTURE_RESOLUTION),8,3);

	scratch=cvLoadImage(STROKE_TEXTURE,1);
	if (!scratch) {
		printf("Failed to load stroke texture %s\n",STROKE_TEXTURE);
	}
	greytmp=cvCreateImage(cvSize(scratch->width,scratch->height),8,1);
	if(scratch->nChannels ==3)
		cvCvtColor( scratch,greytmp, CV_RGB2GRAY);	

	scratch=cvLoadImage(STROKE_BUMP,1);
	if (!scratch) {
		printf("Failed to load stroke normal map (for bump map) %s\n",STROKE_BUMP);
	}
	cvResize(scratch,brushTexBump,CV_INTER_CUBIC);
	cvResize(greytmp,brushTexMask,CV_INTER_CUBIC);
	cvReleaseImage(&scratch);
	cvReleaseImage(&greytmp);


	cvSet(brushTexOpacity,cvScalarAll(255),NULL);	
	uploadTexture(brushTexOpacity, brushTexMask, 0);
	cvFlip(brushTexMask,NULL,0);//vertical flip
	uploadTexture(brushTexOpacity, brushTexMask, 1);
	cvFlip(brushTexMask,NULL,0);//vertical flip
	cvFlip(brushTexMask,NULL,1);//horiz flip
	uploadTexture(brushTexOpacity, brushTexMask, 2);
	cvFlip(brushTexMask,NULL,1);//horiz flip
	cvReleaseImage(&brushTexOpacity);


	// multiscale bump maps
	printf("Building MIP Maps\n");
	for (miplevel=(int)(log(TEXTURE_RESOLUTION)/log(2.0)); miplevel>=((int)(log(TEXTURE_RESOLUTION)/log(2.0)) - TEX_MANUALMIP_SUBSAMPLE -1); miplevel --) {
		uploadTexture(brushTexBump, brushTexMask, 4+(((log(TEXTURE_RESOLUTION)/log(2.0))-miplevel)*3));
		cvFlip(brushTexMask,NULL,0);//vertical flip
		cvFlip(brushTexBump,NULL,0);
		uploadTexture(brushTexBump, brushTexMask, 5+(((log(TEXTURE_RESOLUTION)/log(2.0))-miplevel)*3));
		cvFlip(brushTexMask,NULL,0);//vertical flip
		cvFlip(brushTexBump,NULL,0);
		cvFlip(brushTexMask,NULL,1);//horiz flip
		cvFlip(brushTexBump,NULL,1);
		uploadTexture(brushTexBump, brushTexMask, 6+(((log(TEXTURE_RESOLUTION)/log(2.0))-miplevel)*3));
		cvFlip(brushTexMask,NULL,1);//horiz flip
		cvFlip(brushTexBump,NULL,1);
		//subsample
		scratch=cvCreateImage(cvSize(pow(2.0,(double)miplevel),pow(2.0,(double)miplevel)),8,3);
		cvSmooth(brushTexBump,brushTexBump,CV_GAUSSIAN,9,0,0);
		cvResize(brushTexBump,scratch,CV_INTER_CUBIC);
		cvReleaseImage(&brushTexBump);
		brushTexBump=scratch;
		//subsample
		scratch=cvCreateImage(cvSize(pow(2.0,(double)miplevel),pow(2.0,(double)miplevel)),8,1);
		cvResize(brushTexMask,scratch,CV_INTER_CUBIC);
		cvReleaseImage(&brushTexMask);
		brushTexMask=scratch;
	}


	cvReleaseImage(&brushTexBump);
	cvReleaseImage(&brushTexMask);

#ifndef DISABLE_BUMP
	GenNormalizationCubeMap(128, textureNames[3]);		// Generate Cubemap with a size of 128x128 each face
	SetUp_GL_ARB_multitexture();								// Setting up Multitexturing
#endif

}

void GLreshape ( int w, int h )  {

  CANVAS_HEIGHT=h;
  CANVAS_WIDTH=w;
  
  glViewport(0,0,w,h);
  glutPostRedisplay();

	light_position[0]=IMAGE_WIDTH/2.0;
	light_position[1]=-IMAGE_HEIGHT/2.0;
	light_position[2]=500;//300;
	


}


void GLdisplay(void) {


	float myproj[16];
	float sf,sfx,sfy;

//	glClearColor( 0.35f, 0.53f, 0.7f, 1.0f );
	glClearColor( 0 , 0 ,0 ,1);
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable (GL_TEXTURE_2D); /* enable texture mapping */
	glDisable( GL_DEPTH_TEST );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );



  // Ortho projection
  glMatrixMode   ( GL_PROJECTION );             // Select The Projection Matrix
  glLoadIdentity ( );                // Reset The Projection Matrix


	myproj[0]=1.0/(CANVAS_WIDTH/2.0);
	myproj[1]=0;
	myproj[2]=0;
	myproj[3]=0;

	myproj[4]=0;
	myproj[5]=1.0/(CANVAS_HEIGHT/2.0);
	myproj[6]=0;
	myproj[7]=0;

	myproj[8]=0;
	myproj[9]=0;
	myproj[10]=0;
	myproj[11]=0;

	myproj[12]=0;
	myproj[13]=0;
	myproj[14]=0;
	myproj[15]=1; 

	glLoadMatrixf(myproj); 

  glMatrixMode   ( GL_MODELVIEW );  // Select The Model View Matrix
  glLoadIdentity();

	sfy=((float)CANVAS_HEIGHT)/((float)IMAGE_HEIGHT);
	sfx=((float)CANVAS_WIDTH)/((float)IMAGE_WIDTH);
	sf=min(sfy,sfx);


	glScalef(sf,sf,0);
   glTranslatef(-IMAGE_WIDTH/2.0,-IMAGE_HEIGHT/2.0,0);
  

  plotSplines();


	glutSwapBuffers ( );    // Swap The Buffers To Not Be Left With A Clear Screen
	glFlush();


}


void GLkeyEventSpecial ( int a_keys, int x, int y ) {

  switch ( a_keys ) {
    case GLUT_KEY_UP:     // When Up Arrow Is Pressed...
	  light_position[1]+=10;
      break;

    case GLUT_KEY_DOWN:               // When Down Arrow Is Pressed...
	  light_position[1]-=10;
      break;

    case GLUT_KEY_LEFT:     // When Up Arrow Is Pressed...
	  light_position[0]-=10;
      break;

    case GLUT_KEY_RIGHT:               // When Down Arrow Is Pressed...
	  light_position[0]+=10;
      break;

    default:
      break;

  }
printf("Light pos = %f,%f\n",light_position[0],light_position[1]);
	glutPostRedisplay();

}



void GLkeyEventStandard ( unsigned char key, int x, int y )  {

  switch ( key ) {
    case 27:        // When Escape Is Pressed...
//	  glutReshapeWindow(400,400);
	  freePainting();
	  free(costheta);
	  free(sintheta);
	  free(triangletheta);
	  free(derivtriangletheta);
	  printf("Terminated at user request (escape)\n");
      exit (0);   // Exit The Program
      break;        
    case 'F':        // When 'F' Is Pressed...		
	case 'f':
		if (isFullScreen) {
			glutReshapeWindow(400,400);
			isFullScreen=0;
		}
		else {
		    glutFullScreen ( ); // Go Into Full Screen Mode
			isFullScreen=1;			
		}
		break;
    default:        
      break;
  }

}

void GLidleDispatch (void) {

	
}



void uploadTexture(IplImage* thecanvas, IplImage* alphaimg, int texnum) {

	char* uploadData;
	int i;
	
	uploadData=(char*)calloc((thecanvas->width+2)*(thecanvas->height+2)*4,sizeof(char));	

	for (i=0; i<thecanvas->width*thecanvas->height; i++) {
		*(uploadData+(i*4)+0)=*(thecanvas->imageData+(i*3)+2);
		*(uploadData+(i*4)+1)=*(thecanvas->imageData+(i*3)+0);
		*(uploadData+(i*4)+2)=*(thecanvas->imageData+(i*3)+1);
		*(uploadData+(i*4)+3)=*(alphaimg->imageData+i);
	}


	glBindTexture(GL_TEXTURE_2D,textureNames[texnum]);
	
	glPixelStorei (GL_UNPACK_ALIGNMENT,1);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

#ifdef USE_MIPMAP
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR);
#endif

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	gluBuild2DMipmaps (GL_TEXTURE_2D, 4, thecanvas->width, thecanvas->height, GL_RGBA, GL_UNSIGNED_BYTE, uploadData);

	free(uploadData);

}




void freePainting (void) {

	STROKE* curstroke=painting;
	STROKE* tmp;

	if (!curstroke)
		return;

	do {
		tmp=curstroke;
		free(curstroke->spline->pts);
		free(curstroke->spline);
		curstroke=curstroke->next;
		free(tmp);
	} while (curstroke);


}

void plotSplines (void) {


	float u,B0,B1,B2,B3,T0,T1,T2,T3;
	float spline_x,spline_y,spline_tangx,spline_tangy;
	float linear_x,linear_y,linear_tangx,linear_tangy;
	float x,y;
	float t=0;
	float prevt=0;
	float bumpt=0;
	float prevbumpt=0;
	float tangx,tangy,mag;
	int alternate=0;
	int firsttime;
	CvPoint2D32f top;
	CvPoint2D32f bottom;
	CvPoint2D32f prevtop;
	CvPoint2D32f prevbottom;
	float segmentlength,thresh;
	float prevx,prevy;
	float totaltravelled;
	float thisdist;
	float splinetotallength;
	float distancethisseg;
	float prevlinear_x,prevlinear_y;
	float jaggedness,freq,ampl;
	float aheadx,aheady;
	int thetaval,phaseoffset;
	float thetavalF;
	double undtang_x,undtang_y=0;
	float wobbledisp;
	float finalpos_y,finalpos_x;
	float finaltang_y,finaltang_x;
	float orthotangx,orthotangy;
	float shader_red,shader_green,shader_blue;
	int miplevel;
	
	int splinenum;
	STROKE* curstroke=painting;

  	VECTOR3 vertex_to_light;
	VECTOR3 vertex_position;

	if (!curstroke)
		return;

	top.x=top.y=bottom.x=bottom.y=0; // warning eater

	jaggedness=renderparams.jaggedness;
	freq=100.0f + ((1.0f-renderparams.undulation_freq)*100.0f);
	ampl=renderparams.undulation_amp;

#ifndef DISABLE_BUMP
	/* Texture unit 0 - normalisation */
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, textureNames[3]);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_REPLACE) ;
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE) ;
	

	/* Texture unit 1 - bump map */
	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, textureNames[4]);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_DOT3_RGB) ;
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS) ;
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE) ;


	/* Texture unit 2 - colour texture map*/
	glActiveTexture(GL_TEXTURE2);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, textureNames[0]);
	//glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE) ;
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS) ;
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_PRIMARY_COLOR) ;


#endif


	do {		/*** For each spline in the painting ***/


		firsttime=1;	// don't plot triangles for the very first bit of spline
		phaseoffset=(int)floor(curstroke->ERM*180.0f);

		shader_red=(float)curstroke->red/255.0;
		shader_green=(float)curstroke->green/255.0;
		shader_blue=(float)curstroke->blue/255.0;



		// decide level of detail for stroke texture via thickness (stroke will have tex height 2*thickness)
		miplevel=(int)ceil(log(curstroke->thickness*1.5)/log(2.0));  // yes it should be thickness*2 but i've reduced it
		miplevel=((int)(log(TEXTURE_RESOLUTION)/log(2.0)))-miplevel;
		miplevel=max(0,miplevel);
		miplevel=min(miplevel,TEX_MANUALMIP_SUBSAMPLE);




		glColor3f(shader_red,shader_green,shader_blue);
#ifdef DISABLE_BUMP
		switch ((int)floor(curstroke->ERM*3.0f)) {
			case 0:
			glBindTexture(GL_TEXTURE_2D,textureNames[0]);
			break;
			case 1:
			glBindTexture(GL_TEXTURE_2D,textureNames[1]);
			break;
			case 2:
			glBindTexture(GL_TEXTURE_2D,textureNames[2]);
			break;
		}
#else 
		switch ((int)floor(curstroke->ERM *3.0f)) {
			case 0:
			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, textureNames[4+(miplevel*3)]);			
			glActiveTexture(GL_TEXTURE2);
			glBindTexture(GL_TEXTURE_2D, textureNames[0]);
			break;
			case 1:
			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, textureNames[5+(miplevel*3)]);
			glActiveTexture(GL_TEXTURE2);
			glBindTexture(GL_TEXTURE_2D, textureNames[1]);
			break;
			case 2:
			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, textureNames[6+(miplevel*3)]);
			glActiveTexture(GL_TEXTURE2);
			glBindTexture(GL_TEXTURE_2D, textureNames[2]);
			break;
		}
#endif
		glBegin(GL_TRIANGLES);

		splinetotallength=curstroke->spline->totallength;
		totaltravelled=0;
		linear_x=spline_x=curstroke->spline->pts[0].x;
		linear_y=spline_y=curstroke->spline->pts[0].y;

		for (splinenum=0; splinenum<curstroke->spline->numPoints-3; splinenum++) {
			/* for each piecewise segment of the spline */

			segmentlength=curstroke->spline->lengths[splinenum];
			thresh=segmentlength/(float)SPLINETEXTUREINTERVAL;
			distancethisseg=0;

			u=0; // iterate along parametric curve 0,1
			while (1) {

				if (u>1)
					u=1;
				/******** Get spline pos and distance travelled along it *****/
				prevx=spline_x; prevy=spline_y;
				B0 = (-pow(u,3) + 2.0f*pow(u,2) -u )/2.0f;
				B1 = (3.0f*pow(u,3) - 5.0f*pow(u,2) +2.0f)/2.0f;
				B2 = (-3.0f*pow(u,3) +4.0f*pow(u,2) +u )/2.0f;
				B3 = (pow(u,3) - pow(u,2))/2.0f;
				spline_x=curstroke->spline->pts[0+splinenum].x*B0 + 
					curstroke->spline->pts[1+splinenum].x*B1 + 
					curstroke->spline->pts[2+splinenum].x*B2 + 
					curstroke->spline->pts[3+splinenum].x*B3;
				spline_y=curstroke->spline->pts[0+splinenum].y*B0 + 
					curstroke->spline->pts[1+splinenum].y*B1 + 
					curstroke->spline->pts[2+splinenum].y*B2 + 
					curstroke->spline->pts[3+splinenum].y*B3;
				thisdist=sqrt((spline_x-prevx)*(spline_x-prevx) + (spline_y-prevy)*(spline_y-prevy));
					
				thresh-=thisdist;
				distancethisseg+=thisdist;
				totaltravelled+=thisdist;

				/******** Get linear position *****/
				if (u>0) {
					prevlinear_x=linear_x;
					prevlinear_y=linear_y;
				}
				linear_x=curstroke->spline->pts[1+splinenum].x+(distancethisseg/segmentlength)*(curstroke->spline->pts[2+splinenum].x-curstroke->spline->pts[1+splinenum].x);
				linear_y=curstroke->spline->pts[1+splinenum].y+(distancethisseg/segmentlength)*(curstroke->spline->pts[2+splinenum].y-curstroke->spline->pts[1+splinenum].y);

				/** Plot the triangles - if enough distance passed from prev triangles **/
				if (thresh<0 || u<=0 || u>=1) {
					thresh=segmentlength/(float)SPLINETEXTUREINTERVAL;
					prevt=t;
					prevbumpt=bumpt;
#ifndef DO_ARCLENGTH_TEXTUREMAP
					t=(u/(float)(curstroke->spline->numPoints-3)) + ((float)splinenum/(float)(curstroke->spline->numPoints-3));					
#else
					t=(totaltravelled/splinetotallength);
#endif
#ifndef DO_ARCLENGTH_BUMPMAP
					bumpt=(u/(float)(curstroke->spline->numPoints-3)) + ((float)splinenum/(float)(curstroke->spline->numPoints-3));					
#else
					bumpt=(totaltravelled/splinetotallength);
#endif

					/********* Get linear tangent via finite differences *******/

					if (firsttime) {
						/* look ahead for the tangent*/
						aheadx=curstroke->spline->pts[1+splinenum].x+((distancethisseg+SPLINEINCREMENT_HARDCONSTANT)/segmentlength)*(curstroke->spline->pts[2+splinenum].x-curstroke->spline->pts[1+splinenum].x);
						aheady=curstroke->spline->pts[1+splinenum].y+((distancethisseg+SPLINEINCREMENT_HARDCONSTANT)/segmentlength)*(curstroke->spline->pts[2+splinenum].y-curstroke->spline->pts[1+splinenum].y);
						linear_tangx=aheadx-linear_x;
						linear_tangy=aheady-linear_y;
					}	
					else {
						linear_tangx=linear_x-prevlinear_x;
						linear_tangy=linear_y-prevlinear_y;
					}
					mag=sqrt(linear_tangx*linear_tangx + linear_tangy*linear_tangy);
					if (mag==0)
						mag=1;
					linear_tangx=(linear_tangx/mag);
					linear_tangy=(linear_tangy/mag);				

					/********** Get spline tangent via derivatives ********/

					T0= (-3.0f*pow(u,2) + 4.0f*u - 1.0f)/2.0f;
					T1= (9.0f*pow(u,2) - 10.0f*u) /2.0f;
					T2 = (-9.0f*pow(u,2) +8.0f*u +1.0f)/2.0f;
					T3 = (3.0f*pow(u,2) - 2.0f*u)/2.0f;

					spline_tangx=curstroke->spline->pts[0+splinenum].x*T0 + 
						curstroke->spline->pts[1+splinenum].x*T1 + 
						curstroke->spline->pts[2+splinenum].x*T2 + 
						curstroke->spline->pts[3+splinenum].x*T3;
					spline_tangy=curstroke->spline->pts[0+splinenum].y*T0 + 
						curstroke->spline->pts[1+splinenum].y*T1 + 
						curstroke->spline->pts[2+splinenum].y*T2 + 
						curstroke->spline->pts[3+splinenum].y*T3;
					mag=sqrt(spline_tangx*spline_tangx + spline_tangy*spline_tangy);
					if (mag==0)
						mag=1;
					spline_tangx=(spline_tangx/mag);
					spline_tangy=(spline_tangy/mag);

					/********** Get curve/line interpolated position and tangent **********/			

					x=(1-jaggedness)*spline_x + jaggedness*linear_x;
					y=(1-jaggedness)*spline_y + jaggedness*linear_y;
					tangx=(1-jaggedness)*spline_tangx + jaggedness*linear_tangx;
					tangy=(1-jaggedness)*spline_tangy + jaggedness*linear_tangy;

					mag=sqrt(tangx*tangx + tangy*tangy); // normalise
					if (mag==0)
						mag=1;
					tangx=(tangx/mag);
					tangy=(tangy/mag);

					orthotangx=-tangy;
					orthotangy=tangx;
	
					/********* add wobbles and get final top/bottom of stroke ********/
					//thetaval=(((int)floor((totaltravelled/splinetotallength)*freq*(1.0f-(float)curstroke->isBackground*renderparams.dampen_bgstrokes)*360.0))+phaseoffset)%360;
					thetavalF=
						( 
						 (
						  totaltravelled/(freq*(0.8+(curstroke->ERM*0.4)))
						  -floor(
						    totaltravelled/(freq*(0.8+(curstroke->ERM*0.4)))
						  )
						 )  
						*  (1.0f-(float)curstroke->isBackground*renderparams.dampen_bgstrokes)  
						*  360.0f
						)					  
					 +phaseoffset;
					while (thetavalF>360.0f)
						thetavalF-=360.0f;
					while (thetavalF<0)
						thetavalF+=360.0f;	
					thetaval=(int)(floor(thetavalF)*TRIG_LOOKUP_RESOLUTION);

					wobbledisp=(1.0-jaggedness)*sintheta[thetaval] + jaggedness*triangletheta[thetaval];
					wobbledisp*=curstroke->thickness*ampl*((float)(1.0f-(float)curstroke->isBackground*renderparams.dampen_bgstrokes)*1.5f);
//max jitter ampl is 1.5 times the thickness


					undtang_y=(1.0-jaggedness)*costheta[thetaval] + jaggedness*derivtriangletheta[thetaval]; 

					undtang_x=1;
					mag=sqrt(undtang_x*undtang_x + undtang_y*undtang_y); // normalise
					if (mag==0) 
						mag=1;
					undtang_x=(undtang_x/mag);
					undtang_y=(undtang_y/mag);


					// offset along normal
					finalpos_y=y+tangx*wobbledisp;
					finalpos_x=x-tangy*wobbledisp;

					finaltang_x=tangx*undtang_x + orthotangx*undtang_y;
					finaltang_y=tangy*undtang_x + orthotangy*undtang_y;
		

					/*** fix tops and bottmos for thick stroke **/

					finaltang_x*=curstroke->thickness;
					finaltang_y*=curstroke->thickness;

					prevtop=top;
					prevbottom=bottom;
					top.x=finalpos_x+finaltang_y;
					top.y=finalpos_y-finaltang_x;
					bottom.x=finalpos_x-finaltang_y;
					bottom.y=finalpos_y+finaltang_x;

					/***************** NOW PLOT ******************/

					if (!firsttime) {

						vertex_position[0]=prevtop.x, vertex_position[1]=prevtop.y; vertex_position[2]=0;
						vertex_to_light[0]=light_position[0]-vertex_position[0];
						vertex_to_light[1]=light_position[1]-vertex_position[1];
						vertex_to_light[2]=light_position[2]-vertex_position[2];
#ifndef DISABLE_BUMP
						glMultiTexCoord3f(GL_TEXTURE0, vertex_to_light[0], vertex_to_light[1], vertex_to_light[2]);
						glMultiTexCoord2f(GL_TEXTURE1,prevbumpt,1);						
						glMultiTexCoord2f(GL_TEXTURE2,prevt,0);						
#else
						glTexCoord2d(prevt,0);						
#endif
						glVertex3f(vertex_position[0], vertex_position[1], vertex_position[2]);

						vertex_position[0]=prevbottom.x, vertex_position[1]=prevbottom.y; vertex_position[2]=0;
						vertex_to_light[0]=light_position[0]-vertex_position[0];
						vertex_to_light[1]=light_position[1]-vertex_position[1];
						vertex_to_light[2]=light_position[2]-vertex_position[2];
#ifndef DISABLE_BUMP
						glMultiTexCoord3f(GL_TEXTURE0, vertex_to_light[0], vertex_to_light[1], vertex_to_light[2]);
						glMultiTexCoord2f(GL_TEXTURE1,prevbumpt,0);						
						glMultiTexCoord2f(GL_TEXTURE2,prevt,1);						
#else
						glTexCoord2d(prevt,1);						
#endif
						glVertex3f(vertex_position[0], vertex_position[1], vertex_position[2]);

						vertex_position[0]=top.x, vertex_position[1]=top.y; vertex_position[2]=0;
						vertex_to_light[0]=light_position[0]-vertex_position[0];
						vertex_to_light[1]=light_position[1]-vertex_position[1];
						vertex_to_light[2]=light_position[2]-vertex_position[2];
#ifndef DISABLE_BUMP
						glMultiTexCoord3f(GL_TEXTURE0, vertex_to_light[0], vertex_to_light[1], vertex_to_light[2]);
						glMultiTexCoord2f(GL_TEXTURE1,bumpt,1);						
						glMultiTexCoord2f(GL_TEXTURE2,t,0);						
#else
						glTexCoord2d(t,0);						
#endif
						glVertex3f(vertex_position[0], vertex_position[1], vertex_position[2]);

						vertex_position[0]=prevbottom.x, vertex_position[1]=prevbottom.y; vertex_position[2]=0;
						vertex_to_light[0]=light_position[0]-vertex_position[0];
						vertex_to_light[1]=light_position[1]-vertex_position[1];
						vertex_to_light[2]=light_position[2]-vertex_position[2];
#ifndef DISABLE_BUMP
						glMultiTexCoord3f(GL_TEXTURE0, vertex_to_light[0], vertex_to_light[1], vertex_to_light[2]);
						glMultiTexCoord2f(GL_TEXTURE1,prevbumpt,0);						
						glMultiTexCoord2f(GL_TEXTURE2,prevt,1);						
#else
						glTexCoord2d(prevt,1);						
#endif
						glVertex3f(vertex_position[0], vertex_position[1], vertex_position[2]);

						vertex_position[0]=bottom.x, vertex_position[1]=bottom.y; vertex_position[2]=0;
						vertex_to_light[0]=light_position[0]-vertex_position[0];
						vertex_to_light[1]=light_position[1]-vertex_position[1];
						vertex_to_light[2]=light_position[2]-vertex_position[2];
#ifndef DISABLE_BUMP
						glMultiTexCoord3f(GL_TEXTURE0, vertex_to_light[0], vertex_to_light[1], vertex_to_light[2]);
						glMultiTexCoord2f(GL_TEXTURE1,bumpt,0);						
						glMultiTexCoord2f(GL_TEXTURE2,t,1);						
#else
						glTexCoord2d(t,1);						
#endif
						glVertex3f(vertex_position[0], vertex_position[1], vertex_position[2]);

						vertex_position[0]=top.x, vertex_position[1]=top.y; vertex_position[2]=0;
						vertex_to_light[0]=light_position[0]-vertex_position[0];
						vertex_to_light[1]=light_position[1]-vertex_position[1];
						vertex_to_light[2]=light_position[2]-vertex_position[2];
#ifndef DISABLE_BUMP
						glMultiTexCoord3f(GL_TEXTURE0, vertex_to_light[0], vertex_to_light[1], vertex_to_light[2]);
						glMultiTexCoord2f(GL_TEXTURE1,bumpt,1);						
						glMultiTexCoord2f(GL_TEXTURE2,t,0);						
#else
						glTexCoord2d(t,0);						
#endif
						glVertex3f(vertex_position[0], vertex_position[1], vertex_position[2]);

					}
					else {
						firsttime=0;
					}
				
				} // end "if its time to plot a line"
				if (u>=1.0)
					break;
				u+=SPLINEINCREMENT_HARDCONSTANT;
			} // end 'u' iteration 
			
		}	// end for each spline segment
		glEnd(); 

		/* Advance to next in linked list */
		curstroke=curstroke->next;

	} while (curstroke);

}


