/*******************************************\

  OpenGL code for Stroke Based Rendering
  Bump mapped oil paint strokes
  
  (c) John Collomosse 2005

  Created: JPC 22/6/05

  Extracted from "Empathic Painting" NPAR 2006 for 
  EG2011 NPR tutorial (Collomosse and Kyprianidis)

  Released under GPLv3
  http://www.gnu.org/licenses/gpl.html  

\*******************************************/

#include <stdio.h>
#include <time.h>
#include <math.h>

#include "painting.h"

CvRNG		rng_state;			

extern STROKE* painting;
extern int uniqueID;


#define ROOT3 (1.7320508075f)


void zeroPad(IplImage* img, IplImage* result, int padding) {

	// ony for uint8 images

	int x,y;
	unsigned char* in;
	unsigned char* out;

	result->origin=img->origin;

	for (y=0; y<img->height; y++) {
		for (x=0; x<img->width; x++) {
			in=((unsigned char*)img->imageData + x*img->nChannels + (y*img->width*img->nChannels));
			out=((unsigned char*)result->imageData + (x+padding)*result->nChannels + ((y+padding)*result->width*result->nChannels));
			memcpy(out,in,result->nChannels);
		}
	}

}

void zeroUnpad(IplImage* img, IplImage* result, int padding) {

	// ony for uint8 images

	int x,y;
	unsigned char* in;
	unsigned char* out;

	result->origin=img->origin;

	for (y=0; y<result->height; y++) {
		for (x=0; x<result->width; x++) {
			out=((unsigned char*)result->imageData + x*result->nChannels + (y*result->width*img->nChannels));
			in=((unsigned char*)img->imageData + (x+padding)*img->nChannels + ((y+padding)*img->width*img->nChannels));
			memcpy(out,in,result->nChannels);
		}
	}

}

int getArea(IplImage* mask) {

	int area=0;
	int x,y;

	for (x=0; x<mask->width; x++) {
		for (y=0; y<mask->height; y++) {
			if (*( ((unsigned char*)(mask->imageData)) +x+(y*mask->width))) {
				area++;
			}
		}
	}

	return(area);

}



int isPictureEdge(CvPoint2D32f* pts, int numPts) {

	int allhoriz=1;
	int allvert=1;
	int i,tmp;
	float totallen;


	tmp=(int)pts[0].x;
	for (i=0; i<numPts; i++) {
		if (tmp!=pts[i].x)
			allhoriz=0;
	}
	tmp=(int)pts[0].y;
	for (i=0; i<numPts; i++) {
		if (tmp!=pts[i].y)
			allvert=0;
	}

	totallen=0;
	for (i=0; i<numPts-1; i++) {
		totallen+=(float)sqrt(pow(pts[i].x-pts[i+1].x,2)+pow(pts[i].y-pts[i+1].y,2));
	}


	if (allhoriz || allvert && totallen>100)
		return 1;
	else
		return 0;

}


/*********************************************************************************************************/


void buildTestPainting(void) {

	CvPoint2D32f  CP[6];
	

	CP[0].x=100;
	CP[0].y=0;
	CP[1].x=150;
	CP[1].y=100;
	CP[2].x=300;
	CP[2].y=200;
	CP[3].x=450;
	CP[3].y=300;
	CP[4].x=600;
	CP[4].y=500;
	CP[5].x=740;
	CP[5].y=550;
	addStroke(CP,6,255,255,0,15,1,NULL,0,0);

	CP[0].x=10;
	CP[0].y=400;
	CP[1].x=250;
	CP[1].y=400;
	CP[2].x=400;
	CP[2].y=400;
	CP[3].x=650;
	CP[3].y=400;

	addStroke(CP,4,0,255,0,40,1,NULL,0,0);
//return;
	CP[0].x=100;
	CP[0].y=900;
	CP[1].x=150;
	CP[1].y=500;
	CP[2].x=300;
	CP[2].y=200;
	CP[3].x=450;
	CP[3].y=300;
	CP[4].x=500;
	CP[4].y=500;
	CP[5].x=640;
	CP[5].y=550;
	addStroke(CP,6,255,0,255,45,1,NULL,0,0);
}



double grand(void) {

	double r,v1,v2,fac;

  r=2;
  while (r>=1) {
    v1=(2*(cvRandReal(&rng_state))-1);
    v2=(2*(cvRandReal(&rng_state))-1);
    r=v1*v1+v2*v2;
  }
  fac=sqrt(-2*log(r)/r);

  return(v2*fac);

}

double round(double x) {

	if (x>=0.5)
		return (ceil(x));
	else
		return (floor(x));

}



void addStroke(CvPoint2D32f* pts, int numpoints, int r, int g, int b, float thickness, float depth, STROKE* currentlink, int isBackground, float bgcoveragepercent) {

	int i;
	float totallength,dist,prevx,prevy,u,B0,B1,B2,B3,x,y;
	STROKE* newstroke;
	
	newstroke=(STROKE*)calloc(1,sizeof(STROKE));
	if (!currentlink) { // if no entry point specified, then add to start of list
		currentlink=painting;
	}

	// Add new entry to linked list

	newstroke->prev=currentlink;
	if (currentlink!=NULL) {
		newstroke->next=currentlink->next;
		currentlink->next=newstroke;
	}
	else {
		newstroke->next=NULL;
	}

	if (newstroke->next!=NULL) {
		(newstroke->next)->prev=newstroke;
	}

	if (currentlink==NULL) {
		painting=newstroke;
	}

	// Add data to record
	newstroke->red=r;
	newstroke->green=g;
	newstroke->blue=b;
	newstroke->isBackground=isBackground;
	newstroke->bgcoverage=bgcoveragepercent;
	newstroke->ERM=(float)cvRandReal(&rng_state);

	newstroke->spline=(SPLINE*)calloc(1,sizeof(SPLINE));

	newstroke->spline->numPoints=numpoints+2;
	newstroke->spline->pts=(CvPoint2D32f*)calloc(numpoints+2,sizeof(CvPoint2D32f));
	newstroke->spline->lengths=(float*)calloc(numpoints+2,sizeof(float));
	newstroke->id=uniqueID++;
	newstroke->depth=depth;
	newstroke->thickness=thickness/2.0f;

	memcpy((newstroke->spline->pts)+1,pts,numpoints*sizeof(CvPoint2D32f));
	memcpy(newstroke->spline->pts,pts,sizeof(CvPoint2D32f));
	memcpy(newstroke->spline->pts+(numpoints+1),pts+(numpoints-1),sizeof(CvPoint2D32f));

	totallength=0;
	prevx=newstroke->spline->pts[0].x;
	prevy=newstroke->spline->pts[0].y;
	for (i=0; i<newstroke->spline->numPoints-3; i++) {

		dist=0;
		u=0;
		while(1) {
			B0 = (float)((-pow(u,3) + 2.0f*pow(u,2) -u )/2.0f);
			B1 = (float)((3.0f*pow(u,3) - 5.0f*pow(u,2) +2.0f)/2.0f);
			B2 = (float)((-3.0f*pow(u,3) +4.0f*pow(u,2) +u )/2.0f);
			B3 = (float)((pow(u,3) - pow(u,2))/2.0f);
			x=newstroke->spline->pts[0+i].x*B0 + 
				newstroke->spline->pts[1+i].x*B1 + 
				newstroke->spline->pts[2+i].x*B2 + 
				newstroke->spline->pts[3+i].x*B3;
			y=newstroke->spline->pts[0+i].y*B0 + 
				newstroke->spline->pts[1+i].y*B1 + 
				newstroke->spline->pts[2+i].y*B2 + 
				newstroke->spline->pts[3+i].y*B3;
			dist+=(float)sqrt((x-prevx)*(x-prevx) + (y-prevy)*(y-prevy));
			prevx=x;
			prevy=y;
			if (u>=1)
				break;
			u+=(float)SPLINEINCREMENT_HARDCONSTANT;
		}
	
		totallength+=dist;
		newstroke->spline->lengths[i]=dist;

		/*printf("%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f -- %f\n",
			(newstroke->spline->pts+i+0)->x,(newstroke->spline->pts+i+1)->x,(newstroke->spline->pts+i+2)->x,(newstroke->spline->pts+i+3)->x,
			(newstroke->spline->pts+i+0)->y,(newstroke->spline->pts+i+1)->y,(newstroke->spline->pts+i+2)->y,(newstroke->spline->pts+i+3)->y,
			newstroke->spline->lengths[i]);*/
	}
	newstroke->spline->totallength=totallength;

}

void sortPaintingList(void) {


        QSELEM*  thearray;
        int      i,n_entries;
		STROKE* curstroke;

		n_entries=countStrokes();
		if (n_entries==0)
			return;

        /* Quicksort 2 column double array by 1st column */
        thearray=(QSELEM*)calloc(n_entries,sizeof(QSELEM));

		curstroke=painting;
		i=0;
		do {
			thearray[i].sortidx=curstroke->depth;
			thearray[i].ptr=curstroke;
			i++;
			curstroke=curstroke->next;
		} while (curstroke);

        qsort((void*) thearray, (size_t) n_entries, sizeof(QSELEM), comp);

		/* Reconstruct linked list */
		painting=thearray[0].ptr;
		curstroke=thearray[0].ptr;
		for (i=0; i<n_entries; i++) {
			curstroke->next=thearray[i].ptr;
			thearray[i].ptr->prev=curstroke;
			thearray[i].ptr->next=NULL;
			curstroke=thearray[i].ptr;
		}
//exit(0);
        free(thearray);

}

int comp(const void* a, const void* b) {

        if (  ((QSELEM*)a)->sortidx - ((QSELEM*)b)->sortidx <0)
                return (-1);
        else {
                if (((QSELEM*)a)->sortidx - ((QSELEM*)b)->sortidx ==0)
                        return(0);
                else
                        return(1);
        }


}

int countStrokes (void) {

	int ctr=0;
	STROKE* curstroke=painting;

	if (curstroke) {
		do {
			ctr++;
			curstroke=curstroke->next;
		} while (curstroke);
	}

	return(ctr);

}
